package net.jaekl.frank;

import java.io.PrintWriter;
import java.text.DateFormat;
import java.text.MessageFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;
import java.util.TimeZone;

import net.jaekl.frank.octranspo.Route;
import net.jaekl.frank.octranspo.StopInfo;
import net.jaekl.frank.octranspo.Trip;

public class Schedule {
	DateFormat m_hourMinFmt;
	DateFormat m_hourMinSecFmt;
	FrankBundle m_bundle;
	Locale m_locale;
	Style m_style;
	
	public Schedule(Locale locale) {
		m_locale = locale;
		m_style = new Style();
		m_bundle = FrankBundle.getInst(locale);
		m_hourMinFmt = new SimpleDateFormat("hh:mma", locale);
		m_hourMinSecFmt = new SimpleDateFormat("hh:mm:ssa", locale);
		
		// We always display times in Ottawa (local) time, because that's the only 
		// time that makes sense for a user who wants to know when to catch a bus.
		// In the (unlikely?) event that our server is in a different timezone,
		// we need to explicitly specify the time zone that we want to use.
		TimeZone tz = TimeZone.getTimeZone("America/Toronto");
		m_hourMinFmt.setTimeZone(tz);
		m_hourMinSecFmt.setTimeZone(tz);
	}
	
	String trans(String key) {
		return m_bundle.get(key);
	}
	
	String mapUrl(double latitude, double longitude) {
		return "http://www.openstreetmap.org/?mlat=" + latitude + "&mlon=" + longitude + "&zoom=15";
	}

	// Countdown timer that updates time remaining until each bus is expected.
	void writeScript(PrintWriter pw, String remainArray, int remainCount) {
		String min = trans(FrankBundle.MINUTES);
		String sec = trans(FrankBundle.SECONDS);
		
		pw.println("<SCRIPT>");
	    pw.println("var start; var interval; var remain;");
	    pw.println("window.onload = function() { start = new Date(); interval = setInterval(\"Tick()\", 1000); remain = " + remainArray + "; }");
	    pw.println("function Tick() { ");
	    pw.println("  var now = new Date(); var delta = Math.floor((now.getTime() - start.getTime()) / 1000); ");
	    pw.println("  var secs = delta % 60; var mins = Math.floor(delta / 60);");
	    pw.println("  document.getElementById(\"elapsed\").innerHTML = mins + \"" + min + " \" + secs + \"" + sec + "\";");
	    pw.println("  for (var i = 0; i < " + remainCount + "; ++i) {");
	    pw.println("    var id = \"in_\" + i;");
	    pw.println("    var element = document.getElementById(id);");
	    pw.println("    var remain_delta = (remain[i] * 60) + 30 - delta;");
	    pw.println("    if (remain_delta <= 0) {");
	    pw.println("      element.innerHTML = \"----\";");
	    pw.println("    } else {");
	    pw.println("      element.innerHTML = (Math.floor(remain_delta / 60)) + \"" + min + "\";");
	    pw.println("    }");
	    pw.println("  }");
	    pw.println("}");
	    pw.println("</SCRIPT>");

	}
	
	void writeHeader(PrintWriter pw, String title) {
		pw.println("<HTML>");
		pw.println("<HEAD>");
		pw.println("<TITLE>" + title + "</TITLE>");
		new Style().writeStyle(pw);
		pw.println("</HEAD>");
	}
	
	void writePage(PrintWriter pw, StopInfo stopInfo)
	{
		StringBuilder remainArray = new StringBuilder("[ ");
		int           remainCount = 0;
		
		String title = stopInfo.getDescr() + " (" + stopInfo.getStopNo() + ")";
		String min = trans(FrankBundle.MINUTES);
		
		writeHeader(pw, trans(FrankBundle.FRANK) + ": " + title);
		
		pw.println("<BODY>");
		pw.println("  <TABLE ID=\"trips\">");
		pw.println("    <TR><TH COLSPAN=\"5\">" + title + "</TH></TR>");
		pw.println("    <TR><TH>" +
		        trans(FrankBundle.ROUTE) + 
				"</TH><TH>" +
				trans(FrankBundle.DESTINATION) +
				"</TH><TH>" +
				trans(FrankBundle.ETA) +
				"</TH><TH>" +
				trans(FrankBundle.REMAIN) +  
				"</TH><TH>" +
				trans(FrankBundle.GPS_READ) +
				"</TH>");
		
		for (int routeIdx = 0; routeIdx < stopInfo.getNumRoutes(); ++routeIdx) {
			Route route = stopInfo.getRoute(routeIdx);
			for (int tripIdx = 0; tripIdx < route.getNumTrips(); ++tripIdx) {
				Trip trip = route.getTrip(tripIdx);
				boolean isGhost = (trip.getAdjAge() < 0);
				if (isGhost) {
					// GPS is off.  This bus may not exist.
					pw.println("    <TR CLASS=\"ghost\">");
				} else {
					pw.println("    <TR>");
				}
				pw.println("      <TD>" + route.getRouteNo() + "</TD>");
				pw.println("      <TD>" + trip.getDest() + "</TD>");
				pw.println("      <TD>" + m_hourMinFmt.format(trip.getETA()) + "</TD>");
				
				pw.println("      <TD><SPAN ID=\"in_" + remainCount + "\">"+ trip.getAdjTime()+min+"</SPAN></TD>");
				if (remainCount > 0) {
					remainArray.append(", ");
				}
				remainArray.append(trip.getAdjTime());
				remainCount++;
				
				if (isGhost) {
					pw.println("      <TD>" + trans(FrankBundle.GPS_OFF) + "</TD>");
				}
				else {
					pw.println("      <TD><A HREF=\"" + mapUrl(trip.getLatitude(), trip.getLongitude()) + 
							   "\">" + m_hourMinSecFmt.format(trip.getGPSTime()) + "</A></TD>");
				}
				pw.println("    </TR>");
			}
		}
		
		pw.println("  </TABLE>");
		
		String dataCollectedFormat = trans(FrankBundle.DATA_COLLECTED);
		String dataCollected = MessageFormat.format(dataCollectedFormat, m_hourMinSecFmt.format(new Date()));
		pw.println("  <P>" + dataCollected + "</P>");
		
		remainArray.append(" ]");
		writeScript(pw, remainArray.toString(), remainCount);
		pw.println("</BODY></HTML>");
	}
}
