// Copyright (C) 2004, 2014 Christian Jaekl

package net.jaekl.qd.http;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.InvocationTargetException;
import java.nio.charset.Charset;
import java.util.ArrayList;

import net.jaekl.qd.QDException;
import net.jaekl.qd.util.ExceptionUtils;
import net.jaekl.qd.util.InputStreamWrapper;
import net.jaekl.qd.xml.ParseErrorHandler;
import net.jaekl.qd.xml.ParseHandler;
import net.jaekl.qd.xml.ParseResult;

import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.NameValuePair;
import org.apache.http.client.HttpClient;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.impl.client.HttpClientBuilder;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;

public class RequestBroker
{
	final String UTF_8 = "UTF-8";
	final int TIMEOUT_MSEC = 5000;	// Allow at most 5 seconds before we declare the server to be unresponsive
	
	String m_gatewayUrl;
	ArrayList<NameValuePair> m_baseParams;

	public RequestBroker(String gatewayUrl, ArrayList<NameValuePair> baseParams)
	{
		m_gatewayUrl = gatewayUrl;
		m_baseParams = new ArrayList<NameValuePair>(baseParams);
	}
	
	// NB:  Caller is responsible for close()ing the returned InputStream
	//
	InputStream doSubmit(String method, ArrayList<NameValuePair> passedParams) throws QDException
	{
		ArrayList<NameValuePair> params = new ArrayList<NameValuePair>(m_baseParams);
		params.addAll(passedParams);
		
		try {
			RequestConfig.Builder requestBuilder = RequestConfig.custom();
			requestBuilder = requestBuilder.setConnectTimeout(TIMEOUT_MSEC);
			requestBuilder = requestBuilder.setSocketTimeout(TIMEOUT_MSEC);
			HttpClientBuilder builder = HttpClientBuilder.create();
			builder.setDefaultRequestConfig(requestBuilder.build());
			HttpClient httpClient = builder.build(); 
			HttpPost httpPost = new HttpPost(m_gatewayUrl + "/" + method + "/");
			httpPost.setEntity(new UrlEncodedFormEntity(params, UTF_8));
			HttpResponse response = httpClient.execute(httpPost);
			HttpEntity entity = response.getEntity();

			if (null != entity) {
				InputStream is = entity.getContent();
				return is;
			}
		}
		catch (UnsupportedEncodingException uee) {
			// We should actually be guaranteed that this never happens, 
			// because all JVMs are required to support UTF-8
			assert(false);
			throw new Error(uee);
		}
		catch (IOException ioe) {
			throw new QDException(ioe);
		}
		
		return null;
	}

	public String submit(String method, ArrayList<NameValuePair> passedParams) throws QDException
	{
		StringBuilder sb = new StringBuilder();
		InputStream is = null;

		try {
			is = doSubmit(method, passedParams);
			BufferedReader br = new BufferedReader(new InputStreamReader(is));
			String line = br.readLine();
			while (null != line) {
				sb.append(line).append("\n");
				line = br.readLine();
			}
		}
		catch (IOException ioe) {
			throw new QDException(ioe);
		}
		finally {
			ExceptionUtils.tryClose(is);
		}

		return sb.toString();
	}
	
	public ParseResult submitAndParse(String method, 
			  ArrayList<NameValuePair> passedParams, 
			  Class<? extends ParseResult> rootParserClass)
	throws QDException
	{
		return submitAndParse(method, passedParams, rootParserClass, null);
	}
	
	public ParseResult submitAndParse(String method, 
									  ArrayList<NameValuePair> passedParams, 
									  Class<? extends ParseResult> rootParserClass,
									  String rootTagName)
	throws QDException
	{
		ParseResult result = null;
		InputStreamWrapper isw = null;
		Charset utf8 = null;
		
		try {
			utf8 = Charset.forName(UTF_8);
			if (null == rootTagName) {
				result = (ParseResult) rootParserClass.newInstance();
			} 
			else {
				result = (ParseResult) rootParserClass.getDeclaredConstructor(String.class).newInstance(rootTagName);
			}
			isw = new InputStreamWrapper(doSubmit(method, passedParams));
			XMLReader reader = XMLReaderFactory.createXMLReader();
			ParseHandler ph = new ParseHandler(result);
			ParseErrorHandler peh = new ParseErrorHandler();
			reader.setContentHandler(ph);
			reader.setErrorHandler(peh);
			reader.parse(new InputSource(isw));
		} 
		catch ( SAXParseException saxpe ) {
			String response = "<n/a>";
			if (null != isw) {
				response = new String(isw.getHeadBytes(), utf8);
			}
			throw new InvalidResponseException(saxpe, m_gatewayUrl, method, response);
		}
		catch ( InstantiationException
				| InvocationTargetException
				| IllegalAccessException
				| IOException 
				| NoSuchMethodException
				| SAXException 
				e )
		{
			throw new QDException(e);
		} 
		finally {
			ExceptionUtils.tryClose(isw);
		}
		
		return result;
	}
}
