// Copyright (C) 2004, 2014 Christian Jaekl

// Abstract class representing the result of parsing an XML Element.
// A class derived from this one will know how to parse a subtree inside an XML file, and 
// will contain the result of that parse within itself when the parse has completed.
//
// Note that this code will need to be augmented and fixed if XML namespace support is desired.

package net.jaekl.qd.xml;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Stack;

import org.xml.sax.Attributes;

public abstract class ParseResult
{
	Stack<String> m_current;									// Name of the element that we're currently inside
	StringBuilder m_chars;										// character content of m_current.peek()
	ArrayList<ParseResult> m_childParsers;						// Set of all child parsers
	boolean m_haveSeenMyTag;									// Have I encountered my own (root) tag yet?

	String m_tagName;											// Name of the (root) element tag that I'm parsing
	HashSet<String> m_internal;									// Tags that we will store as members of this class instance
	HashMap<String,Class<? extends ParseResult>> m_external;	// Tags that we will store as child ParseResult-derived objects

	@SuppressWarnings("unchecked")
	public ParseResult(String tagName, String[] internalMemberTags, Object[][] externalParserTags)
	{
		m_current = new Stack<String>();
		m_chars = new StringBuilder();
		m_childParsers = new ArrayList<ParseResult>();
		m_haveSeenMyTag = false;
		
		m_tagName = tagName;
		m_internal = new HashSet<String>();
		m_external = new HashMap<String, Class<? extends ParseResult>>();

		for (String internalTag : internalMemberTags) {
			m_internal.add(internalTag);
		}

		for (int idx = 0; idx < externalParserTags.length; ++idx) {
			String externalTag = (String)externalParserTags[idx][0];
			Class<? extends ParseResult>  parserClass = (Class<? extends ParseResult>)externalParserTags[idx][1];
			m_external.put(externalTag, parserClass);
		}
	}

	public abstract void endContents(String uri, String localName, String qName, String chars) throws XmlParseException;
	public abstract void endExternal(String uri, String localName, String qName) throws XmlParseException;
	
	public String getTagName() { return m_tagName; }
	public boolean haveSeenMyTag() { return m_haveSeenMyTag; }

	public void characters(char[] ch, int start, int length) throws XmlParseException
	{
		m_chars.append(ch, start, length);
	}
	
	protected ParseResult[] collectParsedChildren(Class<? extends ParseResult> cls) {
		ArrayList<ParseResult> collection = new ArrayList<ParseResult>();
		Iterator<ParseResult> iter = m_childParsers.iterator();
		while (iter.hasNext()) {
			ParseResult pr = iter.next();
			if (pr.getClass().isAssignableFrom(cls)) {
				collection.add(pr);
				iter.remove();
			}
		}
		
		ParseResult[] result = new ParseResult[collection.size()];
		return collection.toArray(result);
	}

	// returns true if this ParseResult's context has ended with this endElement() call
	public boolean endElement(String uri, String localName, String qName) throws XmlParseException
	{
		assert (null != localName);
		
		boolean isInternal = m_internal.contains(localName);

		if (! m_haveSeenMyTag) {
			// We're in some unrecognised prologue.  Ignore it and move on.
			return false;
		}
		
		if (m_tagName.equals(localName)) {
			validate();
			return true;
		}
		
		if (!isInternal) {
			// Unrecognized tag.  Ignore it.
			return false;
		}
		
		String tag = m_current.pop();
		if ( ! tag.equals(localName) ) {
			throw new MismatchedTagsException(tag, localName);
		}
		
		String chars = m_chars.toString();
		endContents(uri, localName, qName, chars);
		
		return false;
	}
	
	// returns either itself, or a new ParseResult-derived object, whichever should handle parsing the inside of this element
	public ParseResult startElement(String uri, String localName, String qName, Attributes attributes) 
			throws XmlParseException
	{
		assert (null != localName);

		m_chars.setLength(0);
		
		if (! m_haveSeenMyTag) {
			// Have we opened our own (root) tag yet?
			
			if (m_tagName.equals(localName)) {
				m_haveSeenMyTag = true;
				return this;
			}
			else {
				// One of two things has happened here.
				// Either (a) we've got some sort of wrapper here, and have not yet reach our own tag, 
				//     or (b) we're parsing XML that doesn't match expectations.
				// In either case, we're going to ignore this tag, and scan forward looking for our own root.
				return this;
			}
		}

		if (m_internal.contains(localName)) {
			m_current.push(localName);
			return this;
		}

		Class<? extends ParseResult> parserClass = m_external.get(localName);
		if (null != parserClass) {
			try {
				ParseResult childParser = (ParseResult) parserClass.newInstance();
				m_childParsers.add(childParser);
				return childParser.startElement(uri, localName, qName, attributes);
			}
			catch (IllegalAccessException iae) {
				throw new XmlParseException(iae);
			}
			catch (InstantiationException ie) {
				throw new XmlParseException(ie);
			}
		}
		
		// Not a recognized tag.  Ignore it, rather than complaining. 
		return this;
	}
	
	public void validate() throws XmlParseException
	{
		// Default implementation is a no-op.
		// Override if you want to validate on endElement()
	}
}

