package net.jaekl.qd.http;

import java.util.ArrayList;

import org.junit.Assert;
import net.jaekl.qd.QDException;
import net.jaekl.qd.xml.ParseResult;
import net.jaekl.qd.xml.XmlParseException;

import org.apache.http.NameValuePair;
import org.junit.Test;
import org.xml.sax.SAXParseException;

public class RequestBrokerTest {

	private static final String GATEWAY = "http://hostname.bogus/";
	private static final String METHOD = "apiMethod";
	private static final String ROUTE_SUMMARY_FOR_STOP = 
			"<?xml version=\"1.0\" encoding=\"utf-8\"?>\n"
					+ "<soap:Envelope xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n" 
					+  "xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\"\n" 
					+  "xmlns:soap=\"http://schemas.xmlsoap.org/soap/envelope/\">\n"
					+ "	<soap:Body>\n"
					+ "		<GetRouteSummaryForStopResponse xmlns=\"http://octranspo.com\">\n"
					+ "			<GetRouteSummaryForStopResult>\n"
					+ "				<StopNo xmlns=\"http://tempuri.org/\">1234</StopNo>\n"
					+ "				<StopDescription" 
					+ "				 xmlns=\"http://tempuri.org/\">ONE-TWO-THREE-FOUR</StopDescription>\n"
					+ "				<Error xmlns=\"http://tempuri.org/\"/>\n"
					+ "				<Routes xmlns=\"http://tempuri.org/\">\n"
					+ "					<Route>\n"
					+ "						<RouteNo>123</RouteNo>\n"
					+ "						<DirectionID>0</DirectionID>\n"
					+ "						<Direction>NORTH</Direction>\n"
					+ "						<RouteHeading>First Mall</RouteHeading>\n"
					+ "					</Route>\n"
					+ "					<Route>\n"
					+ "						<RouteNo>123</RouteNo>\n"
					+ "						<DirectionID>1</DirectionID>\n"
					+ "						<Direction>SOUTH</Direction>\n"
					+ "						<RouteHeading>Second Mall</RouteHeading>\n"
					+ "					</Route>\n"
					+ "				</Routes>\n"
					+ "			</GetRouteSummaryForStopResult>\n"
					+ "		</GetRouteSummaryForStopResponse>\n"
					+ "	</soap:Body>\n"
					+ "</soap:Envelope>\n";
	
	// Parse sub-tags, handling some internally and some externally
	public static class RouteSummaryParse extends ParseResult {
		private static final String STOP_NO = "StopNo";

		private static final String[] INTERNAL = { STOP_NO };
		private static final Object[][] EXTERNAL = { };
		
		// Data gleaned from parsing
		int m_stopNo;
		
		public RouteSummaryParse() {
			super("GetRouteSummaryForStopResult", INTERNAL, EXTERNAL);
			
			m_stopNo = 0;
		}
		
		public int getStopNo() { return m_stopNo; }

		@Override
		public void endContents(String uri, String localName, String qName,
				String chars) throws XmlParseException 
		{
			if (localName.equals(STOP_NO)) {
				m_stopNo = Integer.parseInt(chars);
			}
		}

		@Override
		public void endExternal(String uri, String localName, String qName)
				throws XmlParseException 
		{
			Assert.fail("Should not be trying to parse external tags here");
		}
	}
	
	// Try submitting a request (with a mocked-out http post component)
	// and validate that we do, in fact, return the expected text (XML).
	@Test
	public void test_submit() throws QDException {
		ArrayList<NameValuePair> emptyParams = new ArrayList<NameValuePair>();
		
		RequestBrokerMock rbm = new RequestBrokerMock(GATEWAY, emptyParams);
		rbm.setResult(METHOD, emptyParams, ROUTE_SUMMARY_FOR_STOP);
		
		String actual = rbm.submit(METHOD, emptyParams);
		
		Assert.assertEquals(ROUTE_SUMMARY_FOR_STOP, actual);
	}
	
	// Try submitting a request (with a mocked-out http post component)
	// and validate that we do, in fact, parse something correctly.
	@Test
	public void test_submitAndParse() throws QDException {
		ArrayList<NameValuePair> emptyParams = new ArrayList<NameValuePair>();
		
		RequestBrokerMock rbm = new RequestBrokerMock(GATEWAY, emptyParams);
		rbm.setResult(METHOD, emptyParams, ROUTE_SUMMARY_FOR_STOP);
		
		ParseResult pr = rbm.submitAndParse(METHOD, emptyParams, RouteSummaryParse.class);
		
		Assert.assertNotNull(pr);
		Assert.assertTrue(pr instanceof RouteSummaryParse);
		RouteSummaryParse rsp = (RouteSummaryParse)pr;
		Assert.assertEquals(1234, rsp.getStopNo());
	}
	
	
	class RequestBrokerParamExpandMock extends RequestBrokerMock {
		private int m_timesCalled;
		private String m_methodPassed;
		private ArrayList<NameValuePair> m_paramsPassed;
		private Class<? extends ParseResult> m_parserClassPassed;
		
		public RequestBrokerParamExpandMock(String gatewayUrl, ArrayList<NameValuePair> baseParams) {
			super(gatewayUrl, baseParams);
			m_timesCalled = 0;
			m_methodPassed = null;
			m_paramsPassed = null;
			m_parserClassPassed = null;
		}
		
		public int getTimesCalled() { return m_timesCalled; }
		public String getMethodPassed() { return m_methodPassed; }
		public ArrayList<NameValuePair> getParamsPassed() { return m_paramsPassed; }
		public Class<? extends ParseResult> getParserClassPassed() { return m_parserClassPassed; }
		
		@Override 
		public ParseResult submitAndParse(String method, 
				  ArrayList<NameValuePair> passedParams, 
				  Class<? extends ParseResult> rootParserClass,
				  String rootTagName)
			 throws QDException
		{
			Assert.assertNull(rootTagName);
			
			m_timesCalled += 1;
			m_methodPassed = method;
			m_paramsPassed = passedParams;
			m_parserClassPassed = rootParserClass;
			
			return null;
		}
	}
	
	// submitAndParse(method, passedParams, rootParser) 
	// should call through to 
	// submitAndParse(method, passedParams, rootParser, rootTagName=null)
	@Test
	public void test_submitAndParse_expandsDefault4thParam() throws QDException {
		final String GATEWAY_URL = "http://bogus.net:1234/gateway";
		final String METHOD = "MyMethod";
		
		RequestBrokerParamExpandMock rbpem = new RequestBrokerParamExpandMock(GATEWAY_URL, new ArrayList<NameValuePair>() );
		ArrayList<NameValuePair> passedParams = new ArrayList<NameValuePair>();
		ParseResult pr = rbpem.submitAndParse(METHOD, passedParams, ParseResult.class);
		
		Assert.assertNull(pr);
		Assert.assertEquals(1, rbpem.getTimesCalled());
		Assert.assertEquals(METHOD, rbpem.getMethodPassed());
		Assert.assertEquals(passedParams, rbpem.getParamsPassed());
		Assert.assertEquals(ParseResult.class, rbpem.getParserClassPassed());
	}
	
	@Test
	public void test_submitAndParse_nonXmlResponse() {
		ArrayList<NameValuePair> emptyParams = new ArrayList<NameValuePair>();
		
		RequestBrokerMock rbm = new RequestBrokerMock(GATEWAY, emptyParams);
		rbm.setResult(METHOD, emptyParams, "No stop number specified");
		
		try {
			rbm.submitAndParse(METHOD, emptyParams, RouteSummaryParse.class);
			Assert.fail("Should have thrown an InvalidResultException");
		}
		catch (QDException qde) {
			Assert.assertTrue(qde instanceof InvalidResponseException);
			Assert.assertTrue(qde.toString().contains(GATEWAY));
			Assert.assertTrue(qde.toString().contains(METHOD));
		}
	}
	
	@Test
	public void test_submitAndParse_throwsInvalidResultException() {
		final String RESPONSE = "Must specify stop number.";
		InvalidResponseException ire = new InvalidResponseException(new SAXParseException("dummy", null), GATEWAY, METHOD, RESPONSE);
		ArrayList<NameValuePair> emptyParams = new ArrayList<NameValuePair>();
		
		RequestBrokerMock rbm = new RequestBrokerMock(GATEWAY, emptyParams);
		rbm.setResult(METHOD, emptyParams, ire);
		
		try {
			rbm.submitAndParse(METHOD, emptyParams, RouteSummaryParse.class);
			Assert.fail("Should have thrown an exception");
		}
		catch (QDException qde) {
			Assert.assertTrue(qde instanceof InvalidResponseException);
			Assert.assertEquals(ire, qde);
			Assert.assertTrue(ire.toString().contains(GATEWAY));
			Assert.assertTrue(ire.toString().contains(METHOD));
			Assert.assertTrue(ire.toString().contains(RESPONSE));
		}
	}
}
